#ifndef __CDCHITMAKER__
#define __CDCHITMAKER__

//*************************************************************************
//* -----------------
//* CDCHitMaker Class
//* -----------------
//* (Description)
//*	This set of classes makes hits.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCCluster:
//*	CDCClusterBuf:
//*	CDCClusterMaker:
//* (Provides)
//*	CDCHitMaker		: steering class
//*	CDCHitBuf		: hit data buffer class
//*	CDCHitLayer		: hit layer class
//*	CDCHitCell		: hit cell class
//*	CDCHitWire		: hit wire class
//*	CDCHit			: hit data class
//*	CDCTimeToXRelation	: x-t relation
//* (Update Record)
//*	1999/12/13  K.Fujii	Original version.
//*	2000/05/14  K.Fujii	Introduced layer/cell/wire structure.
//*************************************************************************

#include "CDCClusterMaker.h"
#include "Lockable.h"

// ---------------------
// X-t Calibration Datum
// ---------------------

class CDCXTDatum : public TObject {
public:
  CDCXTDatum(Int_t n = 0, Double_t *x = 0, Double_t *t = 0);
  CDCXTDatum(CDCXTDatum &xt);
  ~CDCXTDatum();

  void SetDXpar(Int_t n, Double_t *par);

  Int_t WhereIs(Double_t t);
  Double_t X(Double_t t);
  Double_t DX(Double_t x);

private:
  Int_t	    fN;		//! No. data points
  Double_t *fX;		//! x (cm)
  Double_t *fT;		//! t (nsec)
  Int_t     fNpar;	//! No. parameters for DX function
  Double_t *fDXpar;	//! parameters for DX function

  ClassDef(CDCXTDatum, 1) 	// CDC X-T datum
};

// --------------------
// X-t Calibration Data
// --------------------

class CDCTimeToXRelation : public TObjArray {
private:
  class CalibDatum : public TObjArray {
     public:
     	CalibDatum(Int_t last = 999999) : TObjArray(5), fLastRun(last) {}
	~CalibDatum() { if (GetEntries()) Delete(); }

  	Int_t GetLastRun() { return fLastRun; }	
	Double_t X(CDCCluster *clp)
	{
		Int_t n = clp->GetStationNo();
		Int_t a = clp->GetChannelNo();
		Int_t w = ChannelToWire(n,a);
		Double_t t = clp->GetStartTime();
		return ((CDCXTDatum *)At(w))->X(t);
	}
	Double_t DX(CDCCluster *clp)
	{
		Int_t n = clp->GetStationNo();
		Int_t a = clp->GetChannelNo();
		Int_t w = ChannelToWire(n,a);
		return ((CDCXTDatum *)At(w))->DX(X(clp));
	}
     private:
        Int_t ChannelToWire(Int_t n, Int_t a)
        {
           CDCRawData     *rd = (CDCRawData *)gJSF->FindModule("CDCRawData");
           CDCBeginRunBuf *br = rd->GetBeginRunBuf();
           Int_t fm = br->GetHeader()->GetFormatID();
           switch (fm) {
              case 0: {			// 98 format 
                  return a;
              }
              default: {		// new format
                  Int_t id, w; 
                  Int_t cr = 0;	// Only one crate at the moment.
                  br->EChannelToDChannel(cr,n,a,id,w);
                  return w;
              }
           }
        }
     private:
	Int_t	  fLastRun;	//! Last Run Number
  };
  
public:
  CDCTimeToXRelation(Int_t n = 0, CDCXTDatum *xt = 0, Int_t last = 999999);
  CDCTimeToXRelation(Char_t * calibfile);

  virtual ~CDCTimeToXRelation();
  Int_t	   GetLastRun () { return fCurr->GetLastRun();  }	
  Double_t X (CDCCluster *clp) { return fCurr->X(clp);  }
  Double_t DX(CDCCluster *clp) { return fCurr->DX(clp); }

  void Update();
  
private:
  CalibDatum *fCurr;	//! pointer to current calib datum

  ClassDef(CDCTimeToXRelation, 1) 	// CDC Hit Calibration
};

// ---------
// Hit Datum
// ---------
// (Caution)
// 	fWire is set only when you Append this to a CDCHitWire object.
//
class CDCHitWire;

class CDCHit : public TObject, public Lockable {
friend class CDCHitWire;
public:
  CDCHit(CDCCluster *cl = 0);
  CDCHit(CDCHit &ht);
  virtual ~CDCHit();

  inline CDCCluster   *GetCluster() const { return fCluster; }
  inline CDCHitWire   *GetWire   () const { return fWire;    }

  Int_t		GetLayerNo() const;
  Int_t		GetCellNo () const;
  Int_t		GetWireNo () const;
  Double_t 	Xwire     () const;
  Double_t 	Ywire     () const;

  Double_t	Xdrift()  { if (!fDone) Update(); return fXdrift;  }
  Double_t	DXdrift() { if (!fDone) Update(); return fDXdrift; }
  
  Bool_t	IsSortable() const;
  Int_t		Compare(TObject *obj);

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  void		Update();
  void		CalculatePosition();

private:
  Bool_t	fDone;			// (1,0) = (cluster done, not yet)
  CDCCluster   *fCluster;		// pointer to anode cluster
  Double_t	fXdrift;		// drift distance (cm)
  Double_t	fDXdrift;		// error on drift distance (cm)
  CDCHitWire   *fWire;			// pointer to wire this belongs to

  ClassDef(CDCHit, 1) 	// CDC Anode Hit
};

// -------
// HitWire
// -------
class CDCHitCell;

class CDCHitWire : public TObjArray, public Lockable {
friend class CDCHit;
friend class CDCHitCell;
public:
  CDCHitWire(CDCTimeToXRelation *calib = 0);
  virtual ~CDCHitWire();

  inline CDCHitCell *GetCell() const { return fCell;    }

  inline Int_t	  GetLayerNo() const { return fLayerNo; }
  inline Int_t	  GetCellNo () const { return fCellNo;  }
  inline Int_t	  GetWireNo () const { return fWireNo;  }

  inline Double_t X() const { return fXwire; }
  inline Double_t Y() const { return fYwire; }

  Int_t		Append(CDCHit *hit);

  Bool_t	IsSortable() const;
  Int_t		Compare(TObject *obj);

  void 		DebugPrint(const Char_t *opt = "Brief");

  static void ChannelToWire(Int_t dch,
  			    Int_t &l, Int_t &c, Int_t &w, 
  			    Double_t &x, Double_t &y);

private:
  static void ChannelToWire(CDCCluster *cl,
  			    Int_t &l, Int_t &c, Int_t &w, 
  			    Double_t &x, Double_t &y);

private:
  Int_t		fLayerNo;		// layer No.
  Int_t		fCellNo;		// cell No.
  Int_t		fWireNo;		// wire No.
  Double_t	fXwire;			// wire X position (cm)
  Double_t	fYwire;			// wire Y position (cm)
  CDCHitCell   *fCell;			// pointer to cell this belongs to

  CDCTimeToXRelation  *fCalib;		// pointer to Hit calibration data
 
  ClassDef(CDCHitWire, 1) 	// CDC Hit Wire
};
 
// -------
// HitCell
// -------
class CDCHitLayer;

class CDCHitCell : public TObjArray {
friend class CDCHitLayer;
public:
  CDCHitCell();
  virtual ~CDCHitCell();

  inline CDCHitLayer *GetLayer() const { return fLayer;    }

  inline Int_t	  GetLayerNo() const { return fLayerNo; }
  inline Int_t	  GetCellNo () const { return fCellNo;  }

  Int_t		Append(CDCHitWire *wire);

  Bool_t	IsSortable() const;
  Int_t		Compare(TObject *obj);

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  Int_t		fLayerNo;		// layer No.
  Int_t		fCellNo;		// cell No.
  CDCHitLayer  *fLayer;			// pointer to layer this belongs to
 
  ClassDef(CDCHitCell, 1) 	// CDC Hit Cell
};

// --------
// HitLayer
// --------

class CDCHitLayer : public TObjArray {
public:
  CDCHitLayer();
  virtual ~CDCHitLayer();

  inline Int_t	GetLayerNo() const { return fLayerNo; }

  Int_t		Append(CDCHitCell *cell);

  Bool_t	IsSortable() const;
  Int_t		Compare(TObject *obj);

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  Int_t		fLayerNo;		// layer No.
 
  ClassDef(CDCHitLayer, 1) 	// CDC Hit Layer
};

// ----------
// Hit Buffer
// ----------
class CDCHitMaker;

class CDCHitBuf : public JSFEventBuf {
friend class CDCHitMaker;
public:
  CDCHitBuf(CDCTimeToXRelation *calib = 0,
  	       const char *name="CDCHitBuf",
	       const char *title="CDC Hit data", 
	       CDCHitMaker *mod=0);
  CDCHitBuf(CDCHitMaker *mod,
  	       CDCTimeToXRelation *calib,
	       const char *name="CDCHitBuf",
	       const char *title="CDC Hit data");
  virtual ~CDCHitBuf();

  void		SetClusterData(CDCClusterBuf *clbuf);
  void		SetCalibrationData(CDCTimeToXRelation *calib);
  inline void	SetWidthCut(Double_t x) { xWidth = x; }
  inline void	SetPeakCut (Double_t x) { xPeak  = x; }
  inline void	SetSumCut  (Double_t x) { xSum   = x; }

  Bool_t	IsInitialized() const;	// inquires if input data set
  Int_t		GetNlayers();		// returns No. hit layers
  TObjArray    *GetHitLayers();		// returns array of hit layers

  void 		DebugPrint(const Char_t *opt = "Brief");

private:
  void		MakeHits();		// make hits

private:
  Bool_t	      fDone;	 	//  (1,0) = (clustering done, not yet)
  Double_t	      xWidth;		//  width cut
  Double_t	      xPeak;		//  peak cut
  Double_t	      xSum;		//  sum cut
  CDCClusterBuf	     *fClusterData;	//  Array of clusters
  TObjArray	      fHitLayers; 	//  array of hit layers
  CDCTimeToXRelation *fCalib;		//  pointer to Hit calibration data

  ClassDef(CDCHitBuf, 1) // CDC hit data buffer
};


// ---------
// Hit Maker
// ---------

class CDCHitMaker : public JSFModule {
public:
  CDCHitMaker(CDCTimeToXRelation *calib = 0,
  		 const Char_t *name="CDCHitMaker",
		 const Char_t *title="CDC Hit Maker");
  virtual ~CDCHitMaker();

  void 		SetCalibrationData(CDCTimeToXRelation *calib);

  Bool_t 	Initialize();		// job initialization
  Bool_t 	BeginRun(Int_t run);	// process Begin Run record
  Bool_t 	Process(Int_t ev);	// process event

private:
  Bool_t	      fToDeleteCalib;	// delete Calib in dtor or not
  CDCTimeToXRelation  *fCalib;		// pointer to Hit calibration data
  
  ClassDef(CDCHitMaker, 1) 	// CDC Hit Maker
};

#endif








